@@ -8,7 +8,7 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2002-2013 Nicholas Nethercote
+   Copyright (C) 2002-2017 Nicholas Nethercote
       njn@valgrind.org
 
    This program is free software; you can redistribute it and/or
@@ -22,9 +22,7 @@
    General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
-   02111-1307, USA.
+   along with this program; if not, see <http://www.gnu.org/licenses/>.
 
    The GNU General Public License is contained in the file COPYING.
 */
@@ -538,9 +536,11 @@ get_cache_info(VexArchInfo *vai)
    return ret == 0 ? True : False;
 }
 
-#elif defined(VGA_arm) || defined(VGA_ppc32) || defined(VGA_ppc64) || \
-   defined(VGA_mips32) || defined(VGA_mips64) || defined(VGA_arm64)
-
+#elif defined(VGA_arm) || defined(VGA_ppc32)    || \
+   defined(VGA_ppc64be) || defined(VGA_ppc64le) || \
+   defined(VGA_mips32) || defined(VGA_mips64) || \
+   defined(VGA_arm64) || defined(VGA_nanomips) || \
+   defined(VGA_riscv64)
 static Bool
 get_cache_info(VexArchInfo *vai)
 {
