@@ -8,7 +8,7 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2008-2013 OpenWorks LLP
+   Copyright (C) 2008-2017 OpenWorks LLP
       info@open-works.co.uk
 
    This program is free software; you can redistribute it and/or
@@ -22,9 +22,7 @@
    General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
-   02111-1307, USA.
+   along with this program; if not, see <http://www.gnu.org/licenses/>.
 
    The GNU General Public License is contained in the file COPYING.
 
@@ -130,6 +128,16 @@ const HChar* ML_(pp_DW_TAG) ( DW_TAG tag )
       case DW_TAG_type_unit:          return "DW_TAG_type_unit";
       case DW_TAG_rvalue_reference_type: return "DW_TAG_rvalue_reference_type";
       case DW_TAG_template_alias:     return "DW_TAG_template_alias";
+      /* DWARF 5.  */
+      case DW_TAG_coarray_type:       return "DW_TAG_coarray_type";
+      case DW_TAG_generic_subrange:   return "DW_TAG_generic_subrange";
+      case DW_TAG_dynamic_type:       return "DW_TAG_dynamic_type";
+      case DW_TAG_atomic_type:        return "DW_TAG_atomic_type";
+      case DW_TAG_call_site:          return "DW_TAG_call_site";
+      case DW_TAG_call_site_parameter:
+         return "DW_TAG_call_site_parameter";
+      case DW_TAG_skeleton_unit:      return "DW_TAG_skeleton_unit";
+      case DW_TAG_immutable_type:     return "DW_TAG_immutable_type";
       /* SGI/MIPS Extensions.  */
       case DW_TAG_MIPS_loop:          return "DW_TAG_MIPS_loop";
       /* HP extensions.  See:
@@ -142,6 +150,16 @@ const HChar* ML_(pp_DW_TAG) ( DW_TAG tag )
       case DW_TAG_class_template:     return "DW_TAG_class_template";
       case DW_TAG_GNU_BINCL:          return "DW_TAG_GNU_BINCL";
       case DW_TAG_GNU_EINCL:          return "DW_TAG_GNU_EINCL";
+      case DW_TAG_GNU_template_template_param:
+         return "DW_TAG_GNU_template_template_param";
+      case DW_TAG_GNU_template_parameter_pack:
+         return"DW_TAG_GNU_template_parameter_pack";
+      case DW_TAG_GNU_formal_parameter_pack:
+         return "DW_TAG_GNU_formal_parameter_pack";
+      case DW_TAG_GNU_call_site:
+         return "DW_TAG_GNU_call_site";
+      case DW_TAG_GNU_call_site_parameter:
+         return "DW_TAG_GNU_call_site_parameter";
       /* Extensions for UPC.  See: http://upc.gwu.edu/~upc.  */
       case DW_TAG_upc_shared_type:    return "DW_TAG_upc_shared_type";
       case DW_TAG_upc_strict_type:    return "DW_TAG_upc_strict_type";
@@ -182,6 +200,27 @@ const HChar* ML_(pp_DW_FORM) ( DW_FORM form )
       case DW_FORM_exprloc:   return "DW_FORM_exprloc";
       case DW_FORM_flag_present:return "DW_FORM_flag_present";
       case DW_FORM_ref_sig8:  return "DW_FORM_ref_sig8";
+      case DW_FORM_strx:      return "DW_FORM_strx";
+      case DW_FORM_addrx:     return "DW_FORM_addrx";
+      case DW_FORM_ref_sup4:  return "DW_FORM_ref_sup4";
+      case DW_FORM_strp_sup:  return "DW_FORM_strp_sup";
+      case DW_FORM_data16:    return "DW_FORM_data16";
+      case DW_FORM_line_strp: return "DW_FORM_line_strp";
+      case DW_FORM_implicit_const:return "DW_FORM_implicit_const";
+      case DW_FORM_loclistx:  return "DW_FORM_loclistx";
+      case DW_FORM_rnglistx:  return "DW_FORM_rnglistx";
+      case DW_FORM_ref_sup8:  return "DW_FORM_ref_sup8";
+      case DW_FORM_strx1:     return "DW_FORM_strx1";
+      case DW_FORM_strx2:     return "DW_FORM_strx2";
+      case DW_FORM_strx3:     return "DW_FORM_strx3";
+      case DW_FORM_strx4:     return "DW_FORM_strx4";
+      case DW_FORM_addrx1:    return "DW_FORM_addrx1";
+      case DW_FORM_addrx2:    return "DW_FORM_addrx2";
+      case DW_FORM_addrx3:    return "DW_FORM_addrx3";
+      case DW_FORM_addrx4:    return "DW_FORM_addrx4";
+      /* GNU Debug Fission extensions.  */
+      case DW_FORM_GNU_addr_index:return "DW_FORM_GNU_addr_index";
+      case DW_FORM_GNU_str_index:return "DW_FORM_GNU_str_index";
       case DW_FORM_GNU_ref_alt:return "DW_FORM_GNU_ref_alt";
       case DW_FORM_GNU_strp_alt:return "DW_FORM_GNU_strp_alt";
    }
@@ -288,6 +327,36 @@ const HChar* ML_(pp_DW_AT) ( DW_AT attr )
       case DW_AT_const_expr: return "DW_AT_const_expr";
       case DW_AT_enum_class: return "DW_AT_enum_class";
       case DW_AT_linkage_name: return "DW_AT_linkage_name";
+      /* DWARF 5 values.  */
+      case DW_AT_string_length_bit_size: return "DW_AT_string_length_bit_size";
+      case DW_AT_string_length_byte_size: return "DW_AT_string_length_byte_size";
+      case DW_AT_rank: return "DW_AT_rank";
+      case DW_AT_str_offsets_base: return "DW_AT_str_offsets_base";
+      case DW_AT_addr_base: return "DW_AT_addr_base";
+      case DW_AT_rnglists_base: return "DW_AT_rnglists_base";
+      case DW_AT_dwo_name: return "DW_AT_dwo_name";
+      case DW_AT_reference: return "DW_AT_reference";
+      case DW_AT_rvalue_reference: return "DW_AT_rvalue_reference";
+      case DW_AT_macros: return "DW_AT_macros";
+      case DW_AT_call_all_calls: return "DW_AT_call_all_calls";
+      case DW_AT_call_all_source_calls: return "DW_AT_call_all_source_calls";
+      case DW_AT_call_all_tail_calls: return "DW_AT_call_all_tail_calls";
+      case DW_AT_call_return_pc: return "DW_AT_call_return_pc";
+      case DW_AT_call_value: return "DW_AT_call_value";
+      case DW_AT_call_origin: return "DW_AT_call_origin";
+      case DW_AT_call_parameter: return "DW_AT_call_parameter";
+      case DW_AT_call_pc: return "DW_AT_call_pc";
+      case DW_AT_call_tail_call: return "DW_AT_call_tail_call";
+      case DW_AT_call_target: return "DW_AT_call_target";
+      case DW_AT_call_target_clobbered: return "DW_AT_call_target_clobbered";
+      case DW_AT_call_data_location: return "DW_AT_call_data_location";
+      case DW_AT_call_data_value: return "DW_AT_call_data_value";
+      case DW_AT_noreturn: return "DW_AT_noreturn";
+      case DW_AT_alignment: return "DW_AT_alignment";
+      case DW_AT_export_symbols: return "DW_AT_export_symbols";
+      case DW_AT_deleted: return "DW_AT_deleted";
+      case DW_AT_defaulted: return "DW_AT_defaulted";
+      case DW_AT_loclists_base: return "DW_AT_loclists_base";
       /* SGI/MIPS extensions.  */
       /* case DW_AT_MIPS_fde: return "DW_AT_MIPS_fde"; */
       /* DW_AT_MIPS_fde == DW_AT_HP_unmodifiable */
@@ -324,6 +393,36 @@ const HChar* ML_(pp_DW_AT) ( DW_AT attr )
       case DW_AT_body_begin: return "DW_AT_body_begin";
       case DW_AT_body_end: return "DW_AT_body_end";
       case DW_AT_GNU_vector: return "DW_AT_GNU_vector";
+      case DW_AT_GNU_guarded_by: return "DW_AT_GNU_guarded_by";
+      case DW_AT_GNU_pt_guarded_by: return "DW_AT_GNU_pt_guarded_by";
+      case DW_AT_GNU_guarded: return "DW_AT_GNU_guarded";
+      case DW_AT_GNU_pt_guarded: return "DW_AT_GNU_pt_guarded";
+      case DW_AT_GNU_locks_excluded: return "DW_AT_GNU_locks_excluded";
+      case DW_AT_GNU_exclusive_locks_required: return "DW_AT_GNU_exclusive_locks_required";
+      case DW_AT_GNU_shared_locks_required: return "DW_AT_GNU_shared_locks_required";
+      case DW_AT_GNU_odr_signature: return "DW_AT_GNU_odr_signature";
+      case DW_AT_GNU_template_name: return "DW_AT_GNU_template_name";
+      case DW_AT_GNU_call_site_value: return "DW_AT_GNU_call_site_value";
+      case DW_AT_GNU_call_site_data_value: return "DW_AT_GNU_call_site_data_value";
+      case DW_AT_GNU_call_site_target: return "DW_AT_GNU_call_site_target";
+      case DW_AT_GNU_call_site_target_clobbered: return "DW_AT_GNU_call_site_target_clobbered";
+      case DW_AT_GNU_tail_call: return "DW_AT_GNU_tail_call";
+      case DW_AT_GNU_all_tail_call_sites: return "DW_AT_GNU_all_tail_call_sites";
+      case DW_AT_GNU_all_call_sites: return "DW_AT_GNU_all_call_sites";
+      case DW_AT_GNU_all_source_call_sites: return "DW_AT_GNU_all_source_call_sites";
+      case DW_AT_GNU_locviews: return "DW_AT_GNU_locviews";
+      case DW_AT_GNU_entry_view: return "DW_AT_GNU_entry_view";
+      case DW_AT_GNU_macros: return "DW_AT_GNU_macros";
+      case DW_AT_GNU_deleted: return "DW_AT_GNU_deleted";
+      case DW_AT_GNU_dwo_name: return "DW_AT_GNU_dwo_name";
+      case DW_AT_GNU_dwo_id: return "DW_AT_GNU_dwo_id";
+      case DW_AT_GNU_ranges_base: return "DW_AT_GNU_ranges_base";
+      case DW_AT_GNU_addr_base: return "DW_AT_GNU_addr_base";
+      case DW_AT_GNU_pubnames: return "DW_AT_GNU_pubnames";
+      case DW_AT_GNU_pubtypes: return "DW_AT_GNU_pubtypes";
+      case DW_AT_GNU_numerator: return "DW_AT_GNU_numerator";
+      case DW_AT_GNU_denominator: return "DW_AT_GNU_denominator";
+      case DW_AT_GNU_bias: return "DW_AT_GNU_bias";
       /* VMS extensions.  */
       case DW_AT_VMS_rtnbeg_pd_address: return "DW_AT_VMS_rtnbeg_pd_address";
       /* UPC extension.  */
@@ -336,12 +435,43 @@ const HChar* ML_(pp_DW_AT) ( DW_AT attr )
    return "DW_AT_???";
 }
 
+const HChar* ML_(pp_DW_LLE) ( DW_LLE entry )
+{
+   switch (entry) {
+      case DW_LLE_end_of_list: return "DW_LLE_end_of_list";
+      case DW_LLE_base_addressx: return "DW_LLE_base_addressx";
+      case DW_LLE_startx_endx: return "DW_LLE_startx_endx";
+      case DW_LLE_startx_length: return "DW_LLE_startx_length";
+      case DW_LLE_offset_pair: return "DW_LLE_offset_pair";
+      case DW_LLE_default_location: return "DW_LLE_default_location";
+      case DW_LLE_base_address: return "DW_LLE_base_address";
+      case DW_LLE_start_end: return "DW_LLE_start_end";
+      case DW_LLE_start_length: return "DW_LLE_start_length";
+      case DW_LLE_GNU_view_pair: return "DW_LLE_GNU_view_pair";
+   }
+   return "DW_LLE_???";
+}
+
+const HChar* ML_(pp_DW_RLE) ( DW_RLE entry )
+{
+   switch (entry) {
+      case DW_RLE_end_of_list: return "DW_RLE_end_of_list";
+      case DW_RLE_base_addressx: return "DW_RLE_base_addressx";
+      case DW_RLE_startx_endx: return "DW_RLE_startx_endx";
+      case DW_RLE_startx_length: return "DW_RLE_startx_length";
+      case DW_RLE_offset_pair: return "DW_RLE_offset_pair";
+      case DW_RLE_base_address: return "DW_RLE_base_address";
+      case DW_RLE_start_end: return "DW_RLE_start_end";
+      case DW_RLE_start_length: return "DW_RLE_start_length";
+   }
+   return "DW_RLE_???";
+}
 
 /* ------ To do with evaluation of Dwarf expressions ------ */
 
 /* FIXME: duplicated in readdwarf.c */
 static 
-ULong read_leb128 ( UChar* data, Int* length_return, Int sign )
+ULong read_leb128 ( const UChar* data, Int* length_return, Int sign )
 {
   ULong  result = 0;
   UInt   num_read = 0;
@@ -375,7 +505,7 @@ ULong read_leb128 ( UChar* data, Int* length_return, Int sign )
  * value is returned and the given pointer is
  * moved past end of leb128 data */
 /* FIXME: duplicated in readdwarf.c */
-static ULong read_leb128U( UChar **data )
+static ULong read_leb128U( const UChar **data )
 {
   Int len;
   ULong val = read_leb128( *data, &len, 0 );
@@ -385,7 +515,7 @@ static ULong read_leb128U( UChar **data )
 
 /* Same for signed data */
 /* FIXME: duplicated in readdwarf.c */
-static Long read_leb128S( UChar **data )
+static Long read_leb128S( const UChar **data )
 {
    Int len;
    ULong val = read_leb128( *data, &len, 1 );
@@ -395,18 +525,20 @@ static Long read_leb128S( UChar **data )
 
 /* FIXME: duplicates logic in readdwarf.c: copy_convert_CfiExpr_tree
    and {FP,SP}_REG decls */
-static Bool get_Dwarf_Reg( /*OUT*/Addr* a, Word regno, RegSummary* regs )
+static Bool get_Dwarf_Reg( /*OUT*/Addr* a, Word regno, const RegSummary* regs )
 {
    vg_assert(regs);
-#  if defined(VGP_x86_linux) || defined(VGP_x86_darwin)
+#  if defined(VGP_x86_linux) || defined(VGP_x86_darwin) \
+      || defined(VGP_x86_solaris) || defined(VGP_x86_freebsd)
    if (regno == 5/*EBP*/) { *a = regs->fp; return True; }
    if (regno == 4/*ESP*/) { *a = regs->sp; return True; }
-#  elif defined(VGP_amd64_linux) || defined(VGP_amd64_darwin)
+#  elif defined(VGP_amd64_linux) || defined(VGP_amd64_darwin) \
+        || defined(VGP_amd64_solaris) || defined(VGP_amd64_freebsd)
    if (regno == 6/*RBP*/) { *a = regs->fp; return True; }
    if (regno == 7/*RSP*/) { *a = regs->sp; return True; }
 #  elif defined(VGP_ppc32_linux)
    if (regno == 1/*SP*/) { *a = regs->sp; return True; }
-#  elif defined(VGP_ppc64_linux)
+#  elif defined(VGP_ppc64be_linux) || defined(VGP_ppc64le_linux)
    if (regno == 1/*SP*/) { *a = regs->sp; return True; }
 #  elif defined(VGP_arm_linux)
    if (regno == 13) { *a = regs->sp; return True; }
@@ -414,14 +546,18 @@ static Bool get_Dwarf_Reg( /*OUT*/Addr* a, Word regno, RegSummary* regs )
 #  elif defined(VGP_s390x_linux)
    if (regno == 15) { *a = regs->sp; return True; }
    if (regno == 11) { *a = regs->fp; return True; }
-#  elif defined(VGP_mips32_linux)
+#  elif defined(VGP_mips32_linux) || defined(VGP_nanomips_linux)
    if (regno == 29) { *a = regs->sp; return True; }
    if (regno == 30) { *a = regs->fp; return True; }
 #  elif defined(VGP_mips64_linux)
    if (regno == 29) { *a = regs->sp; return True; }
    if (regno == 30) { *a = regs->fp; return True; }
 #  elif defined(VGP_arm64_linux)
-   I_die_here;
+   if (regno == 31) { *a = regs->sp; return True; }
+   if (regno == 29) { *a = regs->fp; return True; }
+#  elif defined(VGP_riscv64_linux)
+   if (regno == 2) { *a = regs->sp; return True; }
+   if (regno == 8) { *a = regs->fp; return True; }
 #  else
 #    error "Unknown platform"
 #  endif
@@ -471,8 +607,8 @@ static Bool bias_address( Addr* a, const DebugInfo* di )
 
 /* Evaluate a standard DWARF3 expression.  See detailed description in
    priv_d3basics.h.  Doesn't handle DW_OP_piece/DW_OP_bit_piece yet.  */
-GXResult ML_(evaluate_Dwarf3_Expr) ( UChar* expr, UWord exprszB, 
-                                     GExpr* fbGX, RegSummary* regs,
+GXResult ML_(evaluate_Dwarf3_Expr) ( const UChar* expr, UWord exprszB, 
+                                     const GExpr* fbGX, const RegSummary* regs,
                                      const DebugInfo* di,
                                      Bool push_initial_zero )
 {
@@ -504,7 +640,7 @@ GXResult ML_(evaluate_Dwarf3_Expr) ( UChar* expr, UWord exprszB,
       } while (0)
 
    UChar    opcode;
-   UChar*   limit;
+   const UChar* limit;
    Int      sp; /* # of top element: valid is -1 .. N_EXPR_STACK-1 */
    Addr     stack[N_EXPR_STACK]; /* stack of addresses, as per D3 spec */
    GXResult fbval, res;
@@ -726,7 +862,7 @@ GXResult ML_(evaluate_Dwarf3_Expr) ( UChar* expr, UWord exprszB,
             PUSH(uw1);
             break;
          case DW_OP_const1s:
-	    uw1 = *(Char *)expr;
+	    uw1 = *(const Char *)expr;
 	    expr++;
 	    PUSH(uw1);
             break;
@@ -863,7 +999,8 @@ GXResult ML_(evaluate_Dwarf3_Expr) ( UChar* expr, UWord exprszB,
             if (!regs)
                FAIL("evaluate_Dwarf3_Expr: "
                     "DW_OP_call_frame_cfa but no reg info");
-#if defined(VGP_ppc32_linux) || defined(VGP_ppc64_linux)
+#if defined(VGP_ppc32_linux) || defined(VGP_ppc64be_linux) \
+    || defined(VGP_ppc64le_linux)
             /* Valgrind on ppc32/ppc64 currently doesn't use unwind info. */
             uw1 = ML_(read_Addr)((UChar*)regs->sp);
 #else
@@ -914,6 +1051,14 @@ GXResult ML_(evaluate_Dwarf3_Expr) ( UChar* expr, UWord exprszB,
                FAIL("evaluate_Dwarf3_Expr: DW_OP_stack_value "
                     "does not terminate expression");
             break;
+	 case DW_OP_entry_value:
+	 case DW_OP_GNU_entry_value:
+            /* This provides a DWARF expression where any register op
+               needs tobe evaluated as if the value that register had
+               upon entering the function.  Which is non-trivial to
+               implement.  */
+            FAIL("evaluate_Dwarf3_Expr: Unhandled DW_OP entry_value");
+            return res;
          default:
             if (!VG_(clo_xml))
                VG_(message)(Vg_DebugMsg, 
@@ -939,15 +1084,15 @@ GXResult ML_(evaluate_Dwarf3_Expr) ( UChar* expr, UWord exprszB,
 
 /* Evaluate a so-called Guarded (DWARF3) expression.  See detailed
    description in priv_d3basics.h. */
-GXResult ML_(evaluate_GX)( GExpr* gx, GExpr* fbGX,
-                           RegSummary* regs, const DebugInfo* di )
+GXResult ML_(evaluate_GX)( const GExpr* gx, const GExpr* fbGX,
+                           const RegSummary* regs, const DebugInfo* di )
 {
    GXResult res;
    Addr     aMin, aMax;
    UChar    uc;
    UShort   nbytes;
    UWord    nGuards = 0;
-   UChar* p = &gx->payload[0];
+   const UChar* p = &gx->payload[0];
    uc = *p++; /*biasMe*/
    vg_assert(uc == 0 || uc == 1);
    /* in fact it's senseless to evaluate if the guards need biasing.
@@ -966,8 +1111,8 @@ GXResult ML_(evaluate_GX)( GExpr* gx, GExpr* fbGX,
       aMax   = ML_(read_Addr)(p);   p += sizeof(Addr);
       nbytes = ML_(read_UShort)(p); p += sizeof(UShort);
       nGuards++;
-      if (0) VG_(printf)("           guard %d: %#lx %#lx\n",
-                         (Int)nGuards, aMin,aMax);
+      if (0) VG_(printf)("           guard %lu: %#lx %#lx\n",
+                         nGuards, aMin,aMax);
       if (regs == NULL) {
          vg_assert(aMin == (Addr)0);
          vg_assert(aMax == ~(Addr)0);
@@ -1011,7 +1156,7 @@ GXResult ML_(evaluate_GX)( GExpr* gx, GExpr* fbGX,
    Really it ought to be pulled out and turned into a general
    constant- expression evaluator.
 */
-GXResult ML_(evaluate_trivial_GX)( GExpr* gx, const DebugInfo* di )
+GXResult ML_(evaluate_trivial_GX)( const GExpr* gx, const DebugInfo* di )
 {
    GXResult   res;
    Addr       aMin, aMax;
@@ -1021,7 +1166,7 @@ GXResult ML_(evaluate_trivial_GX)( GExpr* gx, const DebugInfo* di )
    MaybeULong *mul, *mul2;
 
    const HChar*  badness = NULL;
-   UChar*  p       = &gx->payload[0]; /* must remain unsigned */
+   const UChar*  p       = &gx->payload[0]; /* must remain unsigned */
    XArray* results = VG_(newXA)( ML_(dinfo_zalloc), "di.d3basics.etG.1",
                                  ML_(dinfo_free),
                                  sizeof(MaybeULong) );
@@ -1110,7 +1255,7 @@ GXResult ML_(evaluate_trivial_GX)( GExpr* gx, const DebugInfo* di )
          VG_(printf)(" ML_(evaluate_trivial_GX): unhandled:\n   ");
          ML_(pp_GX)( gx );
          VG_(printf)("\n");
-         tl_assert(0);
+         vg_assert(0);
       }
       else
          if (!badness)
@@ -1123,10 +1268,10 @@ GXResult ML_(evaluate_trivial_GX)( GExpr* gx, const DebugInfo* di )
 
    res.kind = GXR_Failure;
 
-   tl_assert(nGuards == VG_(sizeXA)( results ));
-   tl_assert(nGuards >= 0);
+   vg_assert(nGuards == VG_(sizeXA)( results ));
+   vg_assert(nGuards >= 0);
    if (nGuards == 0) {
-      tl_assert(!badness);
+      vg_assert(!badness);
       res.word = (UWord)"trivial GExpr has no guards (!)";
       VG_(deleteXA)( results );
       return res;
@@ -1150,11 +1295,11 @@ GXResult ML_(evaluate_trivial_GX)( GExpr* gx, const DebugInfo* di )
    /* All the subexpressions produced a constant, but did they all produce
       the same one? */
    mul = VG_(indexXA)( results, 0 );
-   tl_assert(mul->b == True); /* we just established that all exprs are ok */
+   vg_assert(mul->b == True); /* we just established that all exprs are ok */
 
    for (i = 1; i < nGuards; i++) {
       mul2 = VG_(indexXA)( results, i );
-      tl_assert(mul2->b == True);
+      vg_assert(mul2->b == True);
       if (mul2->ul != mul->ul) {
          res.word = (UWord)"trivial GExpr: subexpressions disagree";
          VG_(deleteXA)( results );
@@ -1188,11 +1333,12 @@ void ML_(pp_GXResult) ( GXResult res )
 }
 
 
-void ML_(pp_GX) ( GExpr* gx ) {
+void ML_(pp_GX) ( const GExpr* gx )
+{
    Addr   aMin, aMax;
    UChar  uc;
    UShort nbytes;
-   UChar* p = &gx->payload[0];
+   const UChar* p = &gx->payload[0];
    uc = *p++;
    VG_(printf)("GX(%s){", uc == 0 ? "final" : "Breqd" );
    vg_assert(uc == 0 || uc == 1);
