@@ -209,7 +209,10 @@ void gdbserver_process_exit_encountered (unsigned char status, Int code)
 static
 const HChar* sym (Addr addr)
 {
-   return VG_(describe_IP) (addr, NULL);
+   // Tracing/debugging so cur_ep is reasonable.
+   const DiEpoch cur_ep = VG_(current_DiEpoch)();
+
+   return VG_(describe_IP) (cur_ep, addr, NULL);
 }
 
 ThreadId vgdb_interrupted_tid = 0;
@@ -287,7 +290,6 @@ void valgrind_resume (struct thread_resume *resume_info)
            "stop_pc %p changed to be resume_pc %s\n",
            C2v(stop_pc), sym(resume_pc));
    }
-   regcache_invalidate();
 }
 
 unsigned char valgrind_wait (char *ourstatus)
@@ -301,7 +303,6 @@ unsigned char valgrind_wait (char *ourstatus)
    pid = VG_(getpid) ();
    dlog(1, "enter valgrind_wait pid %d\n", pid);
 
-   regcache_invalidate();
    valgrind_update_threads(pid);
 
    /* First see if we are done with this process. */
@@ -352,14 +353,13 @@ unsigned char valgrind_wait (char *ourstatus)
 }
 
 /* Fetch one register from valgrind VEX guest state.  */
-static
-void fetch_register (int regno)
+void valgrind_fetch_register (int regno, unsigned char *buf)
 {
    int size;
    ThreadState *tst = (ThreadState *) inferior_target_data (current_inferior);
    ThreadId tid = tst->tid;
 
-   if (regno >= the_low_target.num_regs) {
+   if (regno < 0 || regno >= the_low_target.num_regs) {
       dlog(0, "error fetch_register regno %d max %d\n",
            regno, the_low_target.num_regs);
       return;
@@ -367,16 +367,13 @@ void fetch_register (int regno)
    size = register_size (regno);
    if (size > 0) {
       Bool mod;
-      char buf [size];
       VG_(memset) (buf, 0, size); // registers not fetched will be seen as 0.
       (*the_low_target.transfer_register) (tid, regno, buf,
                                            valgrind_to_gdbserver, size, &mod);
       // Note: the *mod received from transfer_register is not interesting.
-      // We are interested to see if the register data in the register cache is modified.
-      supply_register (regno, buf, &mod);
       if (mod && VG_(debugLog_getLevel)() > 1) {
          char bufimage [2*size + 1];
-         heximage (bufimage, buf, size);
+         heximage (bufimage, (char*) buf, size);
          dlog(3, "fetched register %d size %d name %s value %s tid %u status %s\n", 
               regno, size, the_low_target.reg_defs[regno].name, bufimage, 
               tid, VG_(name_of_ThreadStatus) (tst->status));
@@ -384,102 +381,74 @@ void fetch_register (int regno)
    }
 }
 
-/* Fetch all registers, or just one, from the child process.  */
-static
-void usr_fetch_inferior_registers (int regno)
-{
-   if (regno == -1 || regno == 0)
-      for (regno = 0; regno < the_low_target.num_regs; regno++)
-         fetch_register (regno);
-   else
-      fetch_register (regno);
-}
-
-/* Store our register values back into the inferior.
-   If REGNO is -1, do this for all registers.
-   Otherwise, REGNO specifies which register (so we can save time).  */
-static
-void usr_store_inferior_registers (int regno)
+/* Store register REGNO value back into the inferior VEX state.  */
+void valgrind_store_register (int regno, const unsigned char *buf)
 {
    int size;
    ThreadState *tst = (ThreadState *) inferior_target_data (current_inferior);
    ThreadId tid = tst->tid;
-   
-   if (regno >= 0) {
 
-      if (regno >= the_low_target.num_regs) {
-         dlog(0, "error store_register regno %d max %d\n",
-              regno, the_low_target.num_regs);
-         return;
+   if (regno < 0 || regno >= the_low_target.num_regs) {
+      dlog(0, "error store_register regno %d max %d\n",
+           regno, the_low_target.num_regs);
+      return;
+   }
+
+   size = register_size (regno);
+   if (size > 0) {
+      Bool mod;
+      Addr old_SP, new_SP;
+
+      if (regno == the_low_target.stack_pointer_regno) {
+         /* When the stack pointer register is changed such that
+            the stack is extended, we better inform the tool of the
+            stack increase.  This is needed in particular to avoid
+            spurious Memcheck errors during Inferior calls. So, we
+            save in old_SP the SP before the change. A change of
+            stack pointer is also assumed to have initialised this
+            new stack space. For the typical example of an inferior
+            call, gdb writes arguments on the stack, and then
+            changes the stack pointer. As the stack increase tool
+            function might mark it as undefined, we have to call it
+            at the good moment. */
+         VG_(memset) ((void *) &old_SP, 0, size);
+         (*the_low_target.transfer_register) (tid, regno, (void *) &old_SP,
+                                              valgrind_to_gdbserver, size, &mod);
       }
-      
-      size = register_size (regno);
-      if (size > 0) {
-         Bool mod;
-         Addr old_SP, new_SP;
-         char buf[size];
 
-         if (regno == the_low_target.stack_pointer_regno) {
-            /* When the stack pointer register is changed such that
-               the stack is extended, we better inform the tool of the
-               stack increase.  This is needed in particular to avoid
-               spurious Memcheck errors during Inferior calls. So, we
-               save in old_SP the SP before the change. A change of
-               stack pointer is also assumed to have initialised this
-               new stack space. For the typical example of an inferior
-               call, gdb writes arguments on the stack, and then
-               changes the stack pointer. As the stack increase tool
-               function might mark it as undefined, we have to call it
-               at the good moment. */
-            VG_(memset) ((void *) &old_SP, 0, size);
-            (*the_low_target.transfer_register) (tid, regno, (void *) &old_SP, 
-                                                 valgrind_to_gdbserver, size, &mod);
-         }
+      char buf_copy[size]; 
+      /* copy buf to buf_copy to avoid warnings passing a const to transfer_register.
+         This is ok as transfer_register called with gdbserver_to_valgrind will read from
+         buf and write to VEX state. */
+      VG_(memcpy) (buf_copy, buf, size);
 
-         VG_(memset) (buf, 0, size);
-         collect_register (regno, buf);
-         (*the_low_target.transfer_register) (tid, regno, buf, 
-                                              gdbserver_to_valgrind, size, &mod);
-         if (mod && VG_(debugLog_getLevel)() > 1) {
-            char bufimage [2*size + 1];
-            heximage (bufimage, buf, size);
-            dlog(2, 
-                 "stored register %d size %d name %s value %s "
-                 "tid %u status %s\n", 
-                 regno, size, the_low_target.reg_defs[regno].name, bufimage, 
-                 tid, VG_(name_of_ThreadStatus) (tst->status));
-         }
-         if (regno == the_low_target.stack_pointer_regno) {
-            VG_(memcpy) (&new_SP, buf, size);
-            if (old_SP > new_SP) {
-               Word delta  = (Word)new_SP - (Word)old_SP;
-               dlog(1, 
-                    "   stack increase by stack pointer changed from %p to %p "
-                    "delta %ld\n",
-                    (void*) old_SP, (void *) new_SP,
-                    delta);
-               VG_TRACK( new_mem_stack_w_ECU, new_SP, -delta, 0 );
-               VG_TRACK( new_mem_stack,       new_SP, -delta );
-               VG_TRACK( post_mem_write, Vg_CoreClientReq, tid,
-                         new_SP, -delta);
-            }
+      (*the_low_target.transfer_register) (tid, regno, buf_copy,
+                                           gdbserver_to_valgrind, size, &mod);
+      if (mod && VG_(debugLog_getLevel)() > 1) {
+         char bufimage [2*size + 1];
+         heximage (bufimage, buf_copy, size);
+         dlog(2, 
+              "stored register %d size %d name %s value %s "
+              "tid %u status %s\n", 
+              regno, size, the_low_target.reg_defs[regno].name, bufimage, 
+              tid, VG_(name_of_ThreadStatus) (tst->status));
+      }
+      if (regno == the_low_target.stack_pointer_regno) {
+         VG_(memcpy) (&new_SP, buf, size);
+         if (old_SP > new_SP) {
+            Word delta  = (Word)new_SP - (Word)old_SP;
+            dlog(1, 
+                 "   stack increase by stack pointer changed from %p to %p "
+                 "delta %ld\n",
+                 (void*) old_SP, (void *) new_SP,
+                 delta);
+            VG_TRACK( new_mem_stack_w_ECU, new_SP, -delta, 0 );
+            VG_TRACK( new_mem_stack,       new_SP, -delta );
+            VG_TRACK( post_mem_write, Vg_CoreClientReq, tid,
+                      new_SP, -delta);
          }
       }
    }
-   else {
-      for (regno = 0; regno < the_low_target.num_regs; regno++)
-         usr_store_inferior_registers (regno);
-   }
-}
-
-void valgrind_fetch_registers (int regno)
-{
-   usr_fetch_inferior_registers (regno);
-}
-
-void valgrind_store_registers (int regno)
-{
-   usr_store_inferior_registers (regno);
 }
 
 Bool hostvisibility = False;
@@ -712,13 +681,68 @@ Bool valgrind_get_tls_addr (ThreadState *tst,
    // Check we can read the modid
    CHECK_DEREF(lm+lm_modid_offset, sizeof(unsigned long int), "link_map modid");
    modid = *(unsigned long int *)(lm+lm_modid_offset);
+   dlog (2, "tid %u modid %lu\n", tst->tid, modid);
 
    // Check we can access the dtv entry for modid
    CHECK_DEREF(dtv + 2 * modid, sizeof(CORE_ADDR), "dtv[2*modid]");
 
-   // And finally compute the address of the tls variable.
-   *tls_addr = *(dtv + 2 * modid) + offset;
-   
+   // Compute the base address of the tls block.
+   *tls_addr = *(dtv + 2 * modid);
+
+   if (*tls_addr & 1) {
+      /* This means that computed address is not valid, most probably
+         because given module uses Static TLS.
+         However, the best we can is to try to compute address using
+         static TLS. This is what libthread_db does.
+         Ref. GLIBC/nptl_db/td_thr_tlsbase.c:td_thr_tlsbase().
+      */
+
+      CORE_ADDR tls_offset_addr;
+      PtrdiffT tls_offset;
+
+      dlog(2, "tls_addr (%p & 1) => computing tls_addr using static TLS\n",
+           (void*) *tls_addr);
+
+      /* Assumes that tls_offset is placed right before tls_modid.
+         To check the assumption, start a gdb on none/tests/tls and do:
+           p &((struct link_map*)0x0)->l_tls_modid
+           p &((struct link_map*)0x0)->l_tls_offset
+         Instead of assuming this, we could calculate this similarly to
+         lm_modid_offset, by extending getplatformoffset to support querying
+         more than one offset.
+      */
+      tls_offset_addr = lm + lm_modid_offset - sizeof(PtrdiffT);
+
+      // Check we can read the tls_offset.
+      CHECK_DEREF(tls_offset_addr, sizeof(PtrdiffT), "link_map tls_offset");
+      tls_offset = *(PtrdiffT *)(tls_offset_addr);
+      dlog(2, "tls_offset_addr %p tls_offset %ld\n",
+           (void*)tls_offset_addr, (long)tls_offset);
+
+      /* Following two values represent platform dependent constants
+         NO_TLS_OFFSET and FORCED_DYNAMIC_TLS_OFFSET, respectively. */
+      if ((tls_offset == -1) || (tls_offset == -2)) {
+         dlog(2, "link_map tls_offset is not valid for static TLS\n");
+         return False;
+      }
+
+      // This calculation is also platform dependent.
+#if defined(VGA_mips32) || defined(VGA_mips64)
+      *tls_addr = ((CORE_ADDR)dtv_loc + 2 * sizeof(CORE_ADDR) + tls_offset);
+#elif defined(VGA_ppc64be) || defined(VGA_ppc64le)
+      *tls_addr = ((CORE_ADDR)dtv_loc + sizeof(CORE_ADDR) + tls_offset);
+#elif defined(VGA_x86) || defined(VGA_amd64) || defined(VGA_s390x)
+      *tls_addr = (CORE_ADDR)dtv_loc - tls_offset - sizeof(CORE_ADDR);
+#else
+      // ppc32, arm, arm64
+      dlog(0, "target.c is missing platform code for static TLS\n");
+      return False;
+#endif
+   }
+
+   // Finally, add tls variable offset to tls block base address.
+   *tls_addr += offset;
+
    return True;
 
 #undef CHECK_DEREF
@@ -748,7 +772,6 @@ void initialize_shadow_low(Bool shadow_mode)
     non_shadow_num_regs = the_low_target.num_regs;
   }
 
-  regcache_invalidate();
   if (the_low_target.reg_defs != non_shadow_reg_defs) {
      free (the_low_target.reg_defs);
   }
@@ -842,8 +865,10 @@ void valgrind_initialize_target(void)
    mips32_init_architecture(&the_low_target);
 #elif defined(VGA_mips64)
    mips64_init_architecture(&the_low_target);
-#elif defined(VGA_tilegx)
-   tilegx_init_architecture(&the_low_target);
+#elif defined(VGA_nanomips)
+   nanomips_init_architecture(&the_low_target);
+#elif defined(VGA_riscv64)
+   riscv64_init_architecture(&the_low_target);
 #else
    #error "architecture missing in target.c valgrind_initialize_target"
 #endif
