@@ -8,7 +8,7 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2000-2013 Julian Seward
+   Copyright (C) 2000-2017 Julian Seward
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -22,9 +22,7 @@
    General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
-   02111-1307, USA.
+   along with this program; if not, see <http://www.gnu.org/licenses/>.
 
    The GNU General Public License is contained in the file COPYING.
 */
@@ -52,15 +50,18 @@
 /* Arguments for a syscall. */
 typedef
    struct SyscallArgs {
+#if defined(VGO_freebsd)
+      Word klass;
+#endif
       Word sysno;
-      UWord arg1;
-      UWord arg2;
-      UWord arg3;
-      UWord arg4;
-      UWord arg5;
-      UWord arg6;
-      UWord arg7;
-      UWord arg8;
+      RegWord arg1;
+      RegWord arg2;
+      RegWord arg3;
+      RegWord arg4;
+      RegWord arg5;
+      RegWord arg6;
+      RegWord arg7;
+      RegWord arg8;
    }
    SyscallArgs;
 
@@ -90,9 +91,10 @@ typedef
       // field names), the s_arg value is the offset from the stack pointer.
       Int o_sysno;
 #     if defined(VGP_x86_linux) || defined(VGP_amd64_linux) \
-         || defined(VGP_ppc32_linux) || defined(VGP_ppc64_linux) \
+         || defined(VGP_ppc32_linux) \
          || defined(VGP_arm_linux) || defined(VGP_s390x_linux) \
-         || defined(VGP_mips64_linux) || defined(VGP_arm64_linux)
+         || defined(VGP_arm64_linux) \
+         || defined(VGP_nanomips_linux) || defined(VGP_riscv64_linux)
       Int o_arg1;
       Int o_arg2;
       Int o_arg3;
@@ -101,16 +103,16 @@ typedef
       Int o_arg6;
       Int uu_arg7;
       Int uu_arg8;
-#     elif defined(VGP_mips32_linux)
+#     elif defined(VGP_ppc64be_linux) || defined(VGP_ppc64le_linux)
       Int o_arg1;
       Int o_arg2;
       Int o_arg3;
       Int o_arg4;
-      Int s_arg5;
-      Int s_arg6;
-      Int uu_arg7;
+      Int o_arg5;
+      Int o_arg6;
+      Int o_arg7;
       Int uu_arg8;
-#     elif defined(VGP_x86_darwin)
+#     elif defined(VGP_x86_freebsd)
       Int s_arg1;
       Int s_arg2;
       Int s_arg3;
@@ -119,7 +121,41 @@ typedef
       Int s_arg6;
       Int s_arg7;
       Int s_arg8;
-#     elif defined(VGP_amd64_darwin)
+#     elif defined(VGP_amd64_freebsd)
+      Int o_arg1;
+      Int o_arg2;
+      Int o_arg3;
+      Int o_arg4;
+      Int o_arg5;
+      // arg6 can either be a register or a stack
+      // depending on whether the syscall is 'syscall/__syscall'
+      // or any other syscall
+      union {
+         Int o_arg6;
+         Int s_arg6;
+      };
+      Int s_arg7;
+      Int s_arg8;
+      Bool arg6_is_reg;
+#     elif defined(VGP_mips32_linux)
+      Int o_arg1;
+      Int o_arg2;
+      Int o_arg3;
+      Int o_arg4;
+      Int s_arg5;
+      Int s_arg6;
+      Int s_arg7;
+      Int uu_arg8;
+#     elif defined(VGP_x86_darwin) || defined(VGP_x86_solaris)
+      Int s_arg1;
+      Int s_arg2;
+      Int s_arg3;
+      Int s_arg4;
+      Int s_arg5;
+      Int s_arg6;
+      Int s_arg7;
+      Int s_arg8;
+#     elif defined(VGP_amd64_darwin) || defined(VGP_amd64_solaris)
       Int o_arg1;
       Int o_arg2;
       Int o_arg3;
@@ -128,6 +164,15 @@ typedef
       Int o_arg6;
       Int s_arg7;
       Int s_arg8;
+#     elif defined(VGP_mips64_linux)
+      Int o_arg1;
+      Int o_arg2;
+      Int o_arg3;
+      Int o_arg4;
+      Int o_arg5;
+      Int o_arg6;
+      Int o_arg7;
+      Int o_arg8;
 #     else
 #       error "Unknown platform"
 #     endif
@@ -135,11 +180,12 @@ typedef
    SyscallArgLayout;
 
 /* Flags describing syscall wrappers */
-#define SfMayBlock      (1 << 1) /* may block                         */
-#define SfPostOnFail    (1 << 2) /* call POST() function on failure   */
-#define SfPollAfter     (1 << 3) /* poll for signals on completion    */
-#define SfYieldAfter    (1 << 4) /* yield on completion               */
-#define SfNoWriteResult (1 << 5) /* don't write result to guest state */
+#define SfMayBlock      (1U << 1U) /* may block                         */
+#define SfPostOnFail    (1U << 2U) /* call POST() function on failure   */
+#define SfPollAfter     (1U << 3U) /* poll for signals on completion    */
+#define SfYieldAfter    (1U << 4U) /* yield on completion               */
+#define SfNoWriteResult (1U << 5U) /* don't write result to guest state */
+#define SfKernelRestart (1U << 6U) /* needs a manual restart            */
 
 
 /* ---------------------------------------------------------------------
@@ -179,16 +225,7 @@ typedef
    fixed sized table exposed to the caller, but that's too inflexible;
    hence now use a function which can do arbitrary messing around to
    find the required entry. */
-#if defined(VGP_mips32_linux)
-   /* Up to 6 parameters, 4 in registers 2 on stack. */
-#  define PRA1(s,t,a) PRRAn(1,s,t,a)
-#  define PRA2(s,t,a) PRRAn(2,s,t,a)
-#  define PRA3(s,t,a) PRRAn(3,s,t,a)
-#  define PRA4(s,t,a) PRRAn(4,s,t,a)
-#  define PRA5(s,t,a) PSRAn(5,s,t,a)
-#  define PRA6(s,t,a) PSRAn(6,s,t,a)
 
-#endif
 #if defined(VGO_linux)
 extern
 SyscallTableEntry* ML_(get_linux_syscall_entry)( UInt sysno );
@@ -201,6 +238,14 @@ SyscallTableEntry* ML_(get_linux_syscall_entry)( UInt sysno );
 extern const SyscallTableEntry ML_(syscall_table)[];
 extern const UInt ML_(syscall_table_size);
 
+#elif defined(VGO_solaris)
+extern
+SyscallTableEntry* ML_(get_solaris_syscall_entry)( UInt sysno );
+
+#elif defined(VGO_freebsd)
+extern
+const SyscallTableEntry* ML_(get_freebsd_syscall_entry)( UInt sysno );
+
 #else
 #  error Unknown OS
 #endif   
@@ -229,9 +274,9 @@ extern const UInt ML_(syscall_table_size);
    void vgSysWrap_##auxstr##_##name##_before                     \
                                  ( ThreadId tid,                 \
                                    SyscallArgLayout* layout,     \
-                                   /*MOD*/SyscallArgs* arrghs,   \
-                                   /*OUT*/SyscallStatus* status, \
-                                   /*OUT*/UWord* flags           \
+                                   /*MOD*/ SyscallArgs* arrghs,   \
+                                   /*OUT*/ SyscallStatus* status, \
+                                   /*OUT*/ UWord* flags           \
                                  )
 
 #define DEFN_POST_TEMPLATE(auxstr, name)                         \
@@ -279,7 +324,7 @@ extern const UInt ML_(syscall_table_size);
     vgSysWrap_##auxstr##_##name##_after
 
 /* Add a generic wrapper to a syscall table. */
-#if defined(VGO_linux)
+#if defined(VGO_linux) || defined(VGO_solaris) || defined(VGO_freebsd)
 #  define GENX_(sysno, name)  WRAPPER_ENTRY_X_(generic, sysno, name)
 #  define GENXY(sysno, name)  WRAPPER_ENTRY_XY(generic, sysno, name)
 #elif defined(VGO_darwin)
@@ -294,6 +339,11 @@ extern const UInt ML_(syscall_table_size);
 #define LINX_(sysno, name)    WRAPPER_ENTRY_X_(linux, sysno, name) 
 #define LINXY(sysno, name)    WRAPPER_ENTRY_XY(linux, sysno, name)
 
+/* Add a FreeBSD-specific, arch-independent wrapper to a syscall
+   table. */
+#define BSDX_(sysno, name)    WRAPPER_ENTRY_X_(freebsd, sysno, name) 
+#define BSDXY(sysno, name)    WRAPPER_ENTRY_XY(freebsd, sysno, name)
+
 
 /* ---------------------------------------------------------------------
    Macros useful for writing wrappers concisely.  These refer to the
@@ -313,6 +363,17 @@ extern const UInt ML_(syscall_table_size);
 #define ARG6   (arrghs->arg6)
 #define ARG7   (arrghs->arg7)
 #define ARG8   (arrghs->arg8)
+#define RETVAL2 (arrghs->retval2)
+
+/* Provide signed versions of the argument values */
+#define SARG1  ((Word)ARG1)
+#define SARG2  ((Word)ARG2)
+#define SARG3  ((Word)ARG3)
+#define SARG4  ((Word)ARG4)
+#define SARG5  ((Word)ARG5)
+#define SARG6  ((Word)ARG6)
+#define SARG7  ((Word)ARG7)
+#define SARG8  ((Word)ARG8)
 
 /* Reference to the syscall's current result status/value.  General
    paranoia all round. */
@@ -329,11 +390,13 @@ static inline UWord getRES ( SyscallStatus* st ) {
    return sr_Res(st->sres);
 }
 
+#if defined(VGO_darwin) || defined(VGO_solaris) || defined(VGO_freebsd)
 static inline UWord getRESHI ( SyscallStatus* st ) {
    vg_assert(st->what == SsComplete);
    vg_assert(!sr_isError(st->sres));
    return sr_ResHI(st->sres);
 }
+#endif
 
 static inline UWord getERR ( SyscallStatus* st ) {
    vg_assert(st->what == SsComplete);
@@ -348,6 +411,13 @@ static inline UWord getERR ( SyscallStatus* st ) {
         status->sres = VG_(mk_SysRes_Success)(zzz);  \
    } while (0)
 
+#ifdef VGO_freebsd
+#define SET_STATUS_Success2(zzz, zzz2)               \
+   do { status->what = SsComplete;                   \
+        status->sres = VG_(mk_SysRes_amd64_freebsd)(zzz, zzz2, False);  \
+   } while (0)
+#endif
+
 #define SET_STATUS_Failure(zzz)                      \
    do { Word wzz = (Word)(zzz);                      \
         /* Catch out wildly bogus error values. */   \
@@ -367,8 +437,8 @@ static inline UWord getERR ( SyscallStatus* st ) {
    if (VG_(clo_trace_syscalls))                      \
       VG_(printf)(format, ## args)
 
-#define FUSE_COMPATIBLE_MAY_BLOCK()                       \
-   if (VG_(strstr)(VG_(clo_sim_hints),"fuse-compatible")) \
+#define FUSE_COMPATIBLE_MAY_BLOCK()                            \
+   if (SimHintiS(SimHint_fuse_compatible, VG_(clo_sim_hints))) \
       *flags |= SfMayBlock
 
 
@@ -389,6 +459,7 @@ static inline UWord getERR ( SyscallStatus* st ) {
 #  define PRA4(s,t,a) PRRAn(4,s,t,a)
 #  define PRA5(s,t,a) PSRAn(5,s,t,a)
 #  define PRA6(s,t,a) PSRAn(6,s,t,a)
+#  define PRA7(s,t,a) PSRAn(7,s,t,a)
 
 #elif defined(VGO_linux) && !defined(VGP_mips32_linux)
    /* Up to 6 parameters, all in registers. */
@@ -399,7 +470,7 @@ static inline UWord getERR ( SyscallStatus* st ) {
 #  define PRA5(s,t,a) PRRAn(5,s,t,a)
 #  define PRA6(s,t,a) PRRAn(6,s,t,a)
 
-#elif defined(VGP_x86_darwin)
+#elif defined(VGP_x86_freebsd)
    /* Up to 8 parameters, all on the stack. */
 #  define PRA1(s,t,a) PSRAn(1,s,t,a)
 #  define PRA2(s,t,a) PSRAn(2,s,t,a)
@@ -410,7 +481,36 @@ static inline UWord getERR ( SyscallStatus* st ) {
 #  define PRA7(s,t,a) PSRAn(7,s,t,a)
 #  define PRA8(s,t,a) PSRAn(8,s,t,a)
 
-#elif defined(VGP_amd64_darwin)
+#elif defined(VGP_amd64_freebsd)
+   /* Up to 8 parameters, 6 in registers, 2 on the stack. */
+   /* or 7 in registers and 3 on the stack */
+#  define PRA1(s,t,a) PRRAn(1,s,t,a)
+#  define PRA2(s,t,a) PRRAn(2,s,t,a)
+#  define PRA3(s,t,a) PRRAn(3,s,t,a)
+#  define PRA4(s,t,a) PRRAn(4,s,t,a)
+#  define PRA5(s,t,a) PRRAn(5,s,t,a)
+#  define PRA6(s,t,a) \
+   do { \
+      if (layout->arg6_is_reg) \
+         PRRAn(6,s,t,a); \
+       else \
+         PSRAn(6,s,t,a); \
+   } while (0)
+#  define PRA7(s,t,a) PSRAn(7,s,t,a)
+#  define PRA8(s,t,a) PSRAn(8,s,t,a)
+
+#elif defined(VGP_x86_darwin) || defined(VGP_x86_solaris)
+   /* Up to 8 parameters, all on the stack. */
+#  define PRA1(s,t,a) PSRAn(1,s,t,a)
+#  define PRA2(s,t,a) PSRAn(2,s,t,a)
+#  define PRA3(s,t,a) PSRAn(3,s,t,a)
+#  define PRA4(s,t,a) PSRAn(4,s,t,a)
+#  define PRA5(s,t,a) PSRAn(5,s,t,a)
+#  define PRA6(s,t,a) PSRAn(6,s,t,a)
+#  define PRA7(s,t,a) PSRAn(7,s,t,a)
+#  define PRA8(s,t,a) PSRAn(8,s,t,a)
+
+#elif defined(VGP_amd64_darwin) || defined(VGP_amd64_solaris)
    /* Up to 8 parameters, 6 in registers, 2 on the stack. */
 #  define PRA1(s,t,a) PRRAn(1,s,t,a)
 #  define PRA2(s,t,a) PRRAn(2,s,t,a)
@@ -429,7 +529,7 @@ static inline UWord getERR ( SyscallStatus* st ) {
 /* Tell the tool that the syscall number is being read. */
 #define PRRSN \
       VG_(tdict).track_pre_reg_read(Vg_CoreSysCall, tid, "(syscallno)", \
-                                    layout->o_sysno, sizeof(UWord));
+                                    layout->o_sysno, sizeof(RegWord));
 
 /* REGISTER PARAMETERS */
 
@@ -447,7 +547,7 @@ static inline UWord getERR ( SyscallStatus* st ) {
 #define PRRAn_LE(n,s,t,a)                          \
    do {                                            \
       Int here = layout->o_arg##n;                 \
-      vg_assert(sizeof(t) <= sizeof(UWord));       \
+      vg_assert(sizeof(t) <= sizeof(RegWord));     \
       vg_assert(here >= 0);                        \
       VG_(tdict).track_pre_reg_read(               \
          Vg_CoreSysCall, tid, s"("#a")",           \
@@ -461,16 +561,16 @@ static inline UWord getERR ( SyscallStatus* st ) {
    since the least significant parts of the guest register are stored
    in memory at the highest address.
 */
-#define PRRAn_BE(n,s,t,a)                          \
-   do {                                            \
-      Int here = layout->o_arg##n;                 \
-      Int next = layout->o_arg##n + sizeof(UWord); \
-      vg_assert(sizeof(t) <= sizeof(UWord));       \
-      vg_assert(here >= 0);                        \
-      VG_(tdict).track_pre_reg_read(               \
-         Vg_CoreSysCall, tid, s"("#a")",           \
-         next-sizeof(t), sizeof(t)                 \
-      );                                           \
+#define PRRAn_BE(n,s,t,a)                            \
+   do {                                              \
+      Int here = layout->o_arg##n;                   \
+      Int next = layout->o_arg##n + sizeof(RegWord); \
+      vg_assert(sizeof(t) <= sizeof(RegWord));       \
+      vg_assert(here >= 0);                          \
+      VG_(tdict).track_pre_reg_read(                 \
+         Vg_CoreSysCall, tid, s"("#a")",             \
+         next-sizeof(t), sizeof(t)                   \
+      );                                             \
    } while (0)
 
 #if defined(VG_BIGENDIAN)
@@ -498,7 +598,7 @@ static inline UWord getERR ( SyscallStatus* st ) {
 #define PSRAn_LE(n,s,t,a)                          \
    do {                                            \
       Addr here = layout->s_arg##n + VG_(get_SP)(tid); \
-      vg_assert(sizeof(t) <= sizeof(UWord));       \
+      vg_assert(sizeof(t) <= sizeof(RegWord));     \
       VG_(tdict).track_pre_mem_read(               \
          Vg_CoreSysCallArgInMem, tid, s"("#a")",   \
          here, sizeof(t)                           \
@@ -514,9 +614,9 @@ static inline UWord getERR ( SyscallStatus* st ) {
 #if (defined(VGP_mips32_linux) && defined (_MIPSEB))
  #define PSRAn_BE(n,s,t,a)                                        \
     do {                                                          \
-      Addr next = layout->s_arg##n + sizeof(UWord) +              \
+      Addr next = layout->s_arg##n + sizeof(RegWord) +            \
                   VG_(get_SP)(tid);                               \
-      vg_assert(sizeof(t) <= sizeof(UWord));                      \
+      vg_assert(sizeof(t) <= sizeof(RegWord));                    \
       VG_(tdict).track_pre_mem_read(                              \
          Vg_CoreSysCallArgInMem, tid, s"("#a")",                  \
          next-sizeof(t), sizeof(t)                                \
@@ -525,9 +625,9 @@ static inline UWord getERR ( SyscallStatus* st ) {
 #else
 #define PSRAn_BE(n,s,t,a)                                         \
    do {                                                           \
-      Addr next = layout->o_arg##n + sizeof(UWord) +              \
+      Addr next = layout->o_arg##n + sizeof(RegWord) +            \
                   VG_(threads)[tid].arch.vex.VG_STACK_PTR;        \
-      vg_assert(sizeof(t) <= sizeof(UWord));                      \
+      vg_assert(sizeof(t) <= sizeof(RegWord));                    \
       VG_(tdict).track_pre_mem_read(                              \
          Vg_CoreSysCallArgInMem, tid, s"("#a")",                  \
          next-sizeof(t), sizeof(t)                                \
@@ -619,6 +719,19 @@ static inline UWord getERR ( SyscallStatus* st ) {
 #define POST_FIELD_WRITE(zzfield) \
     POST_MEM_WRITE((UWord)&zzfield, sizeof(zzfield))
 
+// Macros to support 64-bit syscall args split into two 32 bit values
+#define LOHI64(lo,hi)   ( ((ULong)(lo)) | (((ULong)(hi)) << 32) )
+#if defined(VG_LITTLEENDIAN)
+#define MERGE64(lo,hi)   ( ((ULong)(lo)) | (((ULong)(hi)) << 32) )
+#define MERGE64_FIRST(name) name##_low
+#define MERGE64_SECOND(name) name##_high
+#elif defined(VG_BIGENDIAN)
+#define MERGE64(hi,lo)   ( ((ULong)(lo)) | (((ULong)(hi)) << 32) )
+#define MERGE64_FIRST(name) name##_high
+#define MERGE64_SECOND(name) name##_low
+#else
+#error Unknown endianness
+#endif
 
 #endif   // __PRIV_TYPES_N_MACROS_H
 
