@@ -7,7 +7,7 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2000-2013 Julian Seward
+   Copyright (C) 2000-2017 Julian Seward
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -21,9 +21,7 @@
    General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
-   02111-1307, USA.
+   along with this program; if not, see <http://www.gnu.org/licenses/>.
 
    The GNU General Public License is contained in the file COPYING.
 */
@@ -63,22 +61,30 @@ typedef Int ArenaId;
 // This is both the minimum payload size of a malloc'd block, and its
 // minimum alignment.  Must be a power of 2 greater than 4, and should be
 // greater than 8.
-#if   defined(VGP_x86_linux)   || \
-      defined(VGP_arm_linux)   || \
-      defined(VGP_mips32_linux)
+#if   defined(VGP_x86_linux)    || \
+      defined(VGP_arm_linux)    || \
+      defined(VGP_mips32_linux) || \
+      (defined(VGP_mips64_linux) && defined(VGABI_N32)) || \
+      defined(VGP_nanomips_linux) || \
+      defined(VGP_x86_solaris)
 #  define VG_MIN_MALLOC_SZB        8
 // Nb: We always use 16 bytes for Darwin, even on 32-bits, so it can be used
 // for any AltiVec- or SSE-related type.  This matches the Darwin libc.
 // Also, use 16 bytes for any PPC variant, since 16 is required to make
 // Altiveccery work right.
-#elif defined(VGP_amd64_linux)  || \
-      defined(VGP_ppc32_linux)  || \
-      defined(VGP_ppc64_linux)  || \
-      defined(VGP_s390x_linux)  || \
-      defined(VGP_mips64_linux) || \
-      defined(VGP_x86_darwin)   || \
-      defined(VGP_amd64_darwin) || \
-      defined(VGP_arm64_linux)
+#elif defined(VGP_amd64_linux)    || \
+      defined(VGP_ppc32_linux)    || \
+      defined(VGP_ppc64be_linux)  || \
+      defined(VGP_ppc64le_linux)  || \
+      defined(VGP_s390x_linux)    || \
+      (defined(VGP_mips64_linux) && !defined(VGABI_N32)) || \
+      defined(VGP_x86_freebsd)    || \
+      defined(VGP_amd64_freebsd)  || \
+      defined(VGP_x86_darwin)     || \
+      defined(VGP_amd64_darwin)   || \
+      defined(VGP_arm64_linux)    || \
+      defined(VGP_riscv64_linux)  || \
+      defined(VGP_amd64_solaris)
 #  define VG_MIN_MALLOC_SZB       16
 #else
 #  error Unknown platform
@@ -110,6 +116,16 @@ extern void* VG_(arena_memalign)( ArenaId aid, const HChar* cc,
 extern HChar* VG_(arena_strdup)  ( ArenaId aid, const HChar* cc, 
                                    const HChar* s);
 
+/* Specialised version of realloc, that shrinks the size of the block ptr from
+   its current size to req_pszB.
+   req_pszB must be <= to the current size of ptr (otherwise it will assert).
+   Compared to VG_(arena_realloc):
+     * VG_(arena_realloc_shrink) cannot increase the size of ptr.
+     * If large enough, the unused memory is made usable for other allocation.
+     * ptr is shrunk in place, so as to avoid temporary allocation and memcpy. */
+extern void VG_(arena_realloc_shrink) ( ArenaId aid,
+                                        void* ptr, SizeT req_pszB);
+
 extern SizeT VG_(arena_malloc_usable_size) ( ArenaId aid, void* payload );
 
 extern SizeT VG_(arena_redzone_size) ( ArenaId aid );
@@ -127,6 +143,23 @@ extern void  VG_(print_all_arena_stats) ( void );
 
 extern void  VG_(print_arena_cc_analysis) ( void );
 
+typedef 
+   struct _AddrArenaInfo
+   AddrArenaInfo;
+
+struct _AddrArenaInfo {
+   ArenaId aid;
+   const HChar* name; // arena name, !NULL if Addr a points in an arena.
+   SizeT       block_szB;
+   PtrdiffT    rwoffset;
+   Bool        free;  // True if this is in the arena free zone.
+};
+/* If Addr a points in one of the allocation arenas, describes Addr a in *aai
+   otherwise sets *aai to 0/NULL/...
+   Note that no information is produced for addresses allocated with
+   VG_(arena_perm_malloc). */
+extern void VG_(describe_arena_addr) ( Addr a, /*OUT*/AddrArenaInfo* aai );
+
 #endif   // __PUB_CORE_MALLOCFREE_H
 
 /*--------------------------------------------------------------------*/
