@@ -1,7 +1,7 @@
 /*
   This file is part of drd, a thread error detector.
 
-  Copyright (C) 2006-2013 Bart Van Assche <bvanassche@acm.org>.
+  Copyright (C) 2006-2020 Bart Van Assche <bvanassche@acm.org>.
 
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
@@ -14,9 +14,7 @@
   General Public License for more details.
 
   You should have received a copy of the GNU General Public License
-  along with this program; if not, write to the Free Software
-  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
-  02111-1307, USA.
+  along with this program; if not, see <http://www.gnu.org/licenses/>.
 
   The GNU General Public License is contained in the file COPYING.
 */
@@ -43,7 +41,7 @@
 #define STACK_POINTER_OFFSET OFFSET_amd64_RSP
 #elif defined(VGA_ppc32)
 #define STACK_POINTER_OFFSET OFFSET_ppc32_GPR1
-#elif defined(VGA_ppc64)
+#elif defined(VGA_ppc64be) || defined(VGA_ppc64le)
 #define STACK_POINTER_OFFSET OFFSET_ppc64_GPR1
 #elif defined(VGA_arm)
 #define STACK_POINTER_OFFSET OFFSET_arm_R13
@@ -51,10 +49,12 @@
 #define STACK_POINTER_OFFSET OFFSET_arm64_XSP
 #elif defined(VGA_s390x)
 #define STACK_POINTER_OFFSET OFFSET_s390x_r15
-#elif defined(VGA_mips32)
+#elif defined(VGA_mips32) || defined(VGA_nanomips)
 #define STACK_POINTER_OFFSET OFFSET_mips32_r29
 #elif defined(VGA_mips64)
 #define STACK_POINTER_OFFSET OFFSET_mips64_r29
+#elif defined(VGA_riscv64)
+#define STACK_POINTER_OFFSET OFFSET_riscv64_x2
 #else
 #error Unknown architecture.
 #endif
@@ -101,7 +101,7 @@ void DRD_(trace_mem_access)(const Addr addr, const SizeT size,
 
       vc = DRD_(vc_aprint)(DRD_(thread_get_vc)(DRD_(thread_get_running_tid)()));
       if (access_type == eStore && size <= sizeof(HWord)) {
-         DRD_(trace_msg_w_bt)("store 0x%lx size %ld val %ld/0x%lx (thread %d /"
+         DRD_(trace_msg_w_bt)("store 0x%lx size %lu val %lu/0x%lx (thread %u /"
                               " vc %s)", addr, size, stored_value_lo,
                               stored_value_lo, DRD_(thread_get_running_tid)(),
                               vc);
@@ -110,11 +110,11 @@ void DRD_(trace_mem_access)(const Addr addr, const SizeT size,
 
          tl_assert(sizeof(HWord) == 4);
          sv = ((ULong)stored_value_hi << 32) | stored_value_lo;
-         DRD_(trace_msg_w_bt)("store 0x%lx size %ld val %lld/0x%llx (thread %d"
+         DRD_(trace_msg_w_bt)("store 0x%lx size %lu val %llu/0x%llx (thread %u"
                               " / vc %s)", addr, size, sv, sv,
                               DRD_(thread_get_running_tid)(), vc);
       } else {
-         DRD_(trace_msg_w_bt)("%s 0x%lx size %ld (thread %d / vc %s)",
+         DRD_(trace_msg_w_bt)("%s 0x%lx size %lu (thread %u / vc %s)",
                               access_type == eLoad ? "load "
                               : access_type == eStore ? "store"
                               : access_type == eStart ? "start"
@@ -178,7 +178,7 @@ VG_REGPARM(2) void DRD_(trace_load)(Addr addr, SizeT size)
 #ifdef ENABLE_DRD_CONSISTENCY_CHECKS
    /* The assert below has been commented out because of performance reasons.*/
    tl_assert(DRD_(thread_get_running_tid)()
-             == DRD_(VgThreadIdToDrdThreadId)(VG_(get_running_tid())));
+             == DRD_(VgThreadIdToDrdThreadId)(VG_(get_running_tid)()));
 #endif
 
    if (DRD_(running_thread_is_recording_loads)()
@@ -244,7 +244,7 @@ VG_REGPARM(2) void DRD_(trace_store)(Addr addr, SizeT size)
 #ifdef ENABLE_DRD_CONSISTENCY_CHECKS
    /* The assert below has been commented out because of performance reasons.*/
    tl_assert(DRD_(thread_get_running_tid)()
-             == DRD_(VgThreadIdToDrdThreadId)(VG_(get_running_tid())));
+             == DRD_(VgThreadIdToDrdThreadId)(VG_(get_running_tid)()));
 #endif
 
    if (DRD_(running_thread_is_recording_stores)()
@@ -590,9 +590,9 @@ static void instrument_store(IRSB* const bb, IRExpr* addr_expr,
 
 IRSB* DRD_(instrument)(VgCallbackClosure* const closure,
                        IRSB* const bb_in,
-                       VexGuestLayout* const layout,
-                       VexGuestExtents* const vge,
-                       VexArchInfo* archinfo_host,
+                       const VexGuestLayout* const layout,
+                       const VexGuestExtents* const vge,
+                       const VexArchInfo* archinfo_host,
                        IRType const gWordTy,
                        IRType const hWordTy)
 {
@@ -624,7 +624,7 @@ IRSB* DRD_(instrument)(VgCallbackClosure* const closure,
          /* relocated in another way than by later binutils versions. The  */
          /* linker e.g. does not generate .got.plt sections on CentOS 3.0. */
       case Ist_IMark:
-         instrument = VG_(DebugInfo_sect_kind)(NULL, 0, st->Ist.IMark.addr)
+         instrument = VG_(DebugInfo_sect_kind)(NULL, st->Ist.IMark.addr)
             != Vg_SectPLT;
          addStmtToIRSB(bb, st);
          break;
@@ -633,7 +633,9 @@ IRSB* DRD_(instrument)(VgCallbackClosure* const closure,
          switch (st->Ist.MBE.event)
          {
          case Imbe_Fence:
-            break; /* not interesting */
+            break; /* not interesting to DRD */
+         case Imbe_CancelReservation:
+            break; /* not interesting to DRD */
          default:
             tl_assert(0);
          }
